"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit elevations. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import ElevationMarker, FilteredElementCollector,ViewSection

def get_view_index_on_marker(doc, view):
    """
    Get the index of the view on the elevation marker.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: Revit view.
    :type view: Autodesk.Revit.DB.View

    :return: Index of the view on the elevation marker. None if not found.
    :rtype: int or None
    """

    # check if we got an elevation view to start with
    if not isinstance(view, ViewSection):
        return None

    collector = FilteredElementCollector(doc).OfClass(ElevationMarker)

    found_marker = None
    for marker in collector:
        view_count = marker.MaximumViewCount
        for i in range(view_count):
            view_id = marker.GetViewId(i)
            if view_id == view.Id:
                found_marker = i
                break
        
        if found_marker is not None:
            break
    
    return found_marker


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Element filter by design option id's
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from duHast.Revit.Common.design_set_options import (
    get_design_option_ids_of_all_primary_options,
    get_design_option_ids_of_all_primary_options_but_the_one_containing_filter_id,
)
from Autodesk.Revit.DB import Document, ElementId, FilteredElementCollector


def filter_elements_by_design_option_id(
    doc,
    elements,
    design_option_id,
    include_main_model=True,
    include_other_primary=True,
):
    """
    Filter room separation lines by design option id.

    :param doc: The Revit document.
    :type doc: Document
    :param elements: A list of Revit elements
    :type elements: []
    :param design_option_id: The id of the design option to be filtered by. None will return either all room separation lines from the main model, if include main model is true, and / or all room separation lines from other primary design options.
    :type design_option_id: Autodesk.Revit.ElementId
    :param include_main_model: Include room separation lines from the main model option.
    :type include_main_model: bool
    :param include_other_primary: Include room separation lines from other primary design options in the model.
    :type include_other_primary: bool

    :return: A list of room separation lines filtered by design option id.
    :rtype: list
    """

    # do some type checking:
    if not isinstance(doc, Document):
        raise TypeError(
            "doc needs to be of type Document. But got : {}".format(type(doc))
        )
    if not isinstance(elements, list) and not isinstance(
        elements, FilteredElementCollector
    ):
        raise TypeError(
            "elements needs to be of type list or FilteredElementCollector. But got : {}".format(
                type(elements)
            )
        )
    if not isinstance(design_option_id, ElementId) and design_option_id != None:
        raise TypeError(
            "design_option_id needs to be of type ElementId or None. But got : {}".format(
                type(design_option_id)
            )
        )
    if not isinstance(include_main_model, bool):
        raise TypeError(
            "include_main_model needs to be of type bool. But got : {}".format(
                type(include_main_model)
            )
        )
    if not isinstance(include_other_primary, bool):
        raise TypeError(
            "include_other_primary needs to be of type bool. But got : {}".format(
                type(include_other_primary)
            )
        )

    # set up a place holder to hold design option ids to filter by
    option_ids = []
    # get the design option filter ids depending on toggle values and filter id past in
    if (
        design_option_id == None
        and include_main_model == True
        and include_other_primary == True
    ):
        # return lines within the main model and other primary design options
        # get all primary design option ids
        ids_all_primary = get_design_option_ids_of_all_primary_options(doc)
        option_ids.extend(ids_all_primary)
        # add the main model id
        option_ids.append(ElementId.InvalidElementId)
    elif (
        design_option_id == None
        and include_main_model == False
        and include_other_primary == True
    ):
        # return lines within other primary design options only
        # get all primary design option ids
        ids_all_primary = get_design_option_ids_of_all_primary_options(doc)
        option_ids.extend(ids_all_primary)
    elif (
        design_option_id == None
        and include_main_model == True
        and include_other_primary == False
    ):
        # return lines within the main model only
        option_ids.append(ElementId.InvalidElementId)
    elif (
        design_option_id == None
        and include_main_model == False
        and include_other_primary == False
    ):
        # nothing to be returned? Throw an error
        raise ValueError("You need to include at least one option to filter by.")
    elif (
        design_option_id != None
        and include_main_model == True
        and include_other_primary == True
    ):
        # return lines within the specified design option, the main model and other primary design options
        # get all primary design option ids except the one where the design set contains the design option filter id
        ids_primary_filtered = get_design_option_ids_of_all_primary_options_but_the_one_containing_filter_id(
            doc=doc, filter_id=design_option_id
        )
        option_ids.extend(ids_primary_filtered)
        # add the design option filter id
        option_ids.append(design_option_id)
        # add the main model id
        option_ids.append(ElementId.InvalidElementId)
    elif (
        design_option_id != None
        and include_main_model == False
        and include_other_primary == True
    ):
        # return lines within the specified design option and other primary design options only
        # get all primary design option ids except the one where the design set contains the design option filter id
        ids_primary_filtered = get_design_option_ids_of_all_primary_options_but_the_one_containing_filter_id(
            doc=doc, filter_id=design_option_id
        )
        option_ids.extend(ids_primary_filtered)
        # add the design option filter id
        option_ids.append(design_option_id)
    elif (
        design_option_id != None
        and include_main_model == True
        and include_other_primary == False
    ):
        # return lines within the specified design option and the main model only
        # add the design option filter id
        option_ids.append(design_option_id)
        # add the main model id
        option_ids.append(ElementId.InvalidElementId)
    elif (
        design_option_id != None
        and include_main_model == False
        and include_other_primary == False
    ):
        # return lines within the specified design option only
        option_ids.append(design_option_id)

    # set up a return value
    elements_filtered = []

    # loop over lines and filter them by design option id
    for element_instance in elements:
        try:
            # check if the design option id of the element is in the list of design option ids
            element_design_option = element_instance.DesignOption
            # main model is represented by an invalid element id in the design option ;list
            # however on the element its represented by None
            if element_design_option is not None:
                if element_instance.DesignOption.Id in option_ids:
                    elements_filtered.append(element_instance)
            else:
                if ElementId.InvalidElementId in option_ids:
                    elements_filtered.append(element_instance)
        except Exception as e:
            raise ValueError(
                "Error while filtering elements by design option id. Error: {}".format(
                    e
                )
            )

    # return filtered set
    return elements_filtered



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit purging not used shared parameters using purge by delete helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.SharedParameters.shared_parameters import get_all_shared_parameters
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Purge.purge_unused_by_delete import purge_unused_elements

from duHast.Utilities.Objects.result import Result


def get_shared_parameter_ids(doc, element_ids=None, element_ids_list_is_inclusive_filter=True):
    """
    Returns all shared parameter ids in the model with the exception of parameters which have a binding and are therefore assumed to be used.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_ids: optional list of shared parameter element ids
    :type element_ids: [Autodesk.Revit.DB.ElementId]
    :param element_ids_list_is_inclusive_filter: If true and element_ids list has values only those parameters will be purged if possible. If false and element_ids list has values any parameters in the list will not be purged.
    :type element_ids_list_is_inclusive_filter: bool

    :return: A list of all shared parameter ids in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    shared_col = get_all_shared_parameters(doc)

    parameter_with_bindings = []
    # get all parameter binding in a project and exclude those from the list
    # to speed things up a little
    iter = doc.ParameterBindings.ForwardIterator()
    iter.Reset()
    while iter.MoveNext():
        definition = iter.Key
        parameter_with_bindings.append(definition.Id)

    # get the ids of collector
    ids = get_ids_from_element_collector(shared_col)
    # remove all shared parameters which are bound to a category
    ids_without_parameter_binding = [i for i in ids if i not in parameter_with_bindings]

    # check if further filtering is required
    if element_ids == None:
        return ids_without_parameter_binding
    
    # apply filtering
    ids_filtered = []
    if element_ids_list_is_inclusive_filter:
        # only return element ids which are also present in the filter list
        for id_without_binding in ids_without_parameter_binding:
            if id_without_binding in element_ids:
                ids_filtered.append(id_without_binding)
    else:
        # only return element ids which are not present in the filter list
        for id_without_binding in ids_without_parameter_binding:
            if id_without_binding not in element_ids:
                ids_filtered.append(id_without_binding)
    return ids_filtered


def purge_shared_parameters_by_delete(
    doc, progress_callback=None, debug=False, element_ids=None, element_ids_list_is_inclusive_filter=True
):
    """
    Purge shared parameters by delete.

    Note: This is a slow(ish) process and can take a few hours to complete depending on the size of the model and the number of shared parameters.

    No parameters with a category binding will be deleted. (refer below for more details on this)

    Observations:

    - any shared parameter which is bound to a category in a Revit model (but not used in a family etc) will report 2 elements as deleted:
        - the shared parameter itself
        - the binding (type or instance) to a category

    - any shared parameter which is not bound to a category in a Revit model, but used in a family etc will report
        - 1 element as deleted
            - the shared parameter itself
        - multiple changed elements:
            - the families (and any of its instances placed) the shared parameter is used in

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param progress_callback: Callback to report progress.
    :type progress_callback: callable
    :param debug: Debug mode.
    :type debug: bool
    :param element_ids: optional list of shared parameter element ids
    :type element_ids: [Autodesk.Revit.DB.ElementId]
    :param element_ids_list_is_inclusive_filter: If true and element_ids list has values only those parameters will be purged if possible. If false and element_ids list has values any parameters in the list will not be purged.
    :type element_ids_list_is_inclusive_filter: bool

    :return: Result class instance.

        - .status True if unused shared parameters where deleted or nothing needed to be deleted. Otherwise False.
        - .message will contain deletion status.

    """

    return_value = Result()

    try:

        # set up element Id getter
        # make allowance for an ignore element id list
        def action(doc):
            result_action = get_shared_parameter_ids(
                doc=doc, element_ids=element_ids,
                element_ids_list_is_inclusive_filter=element_ids_list_is_inclusive_filter
            )
            return result_action

        # purge unused shared parameters
        purge_result = purge_unused_elements(
            doc=doc,
            element_id_getter=action,
            deleted_elements_modifier=None,
            modified_elements_modifier=None,
            progress_callback=progress_callback,
            debug=debug,
        )

        return_value.update(purge_result)
    except Exception as e:
        return_value.update_sep(
            False, "Error purging shared parameters by delete: {}".format(str(e))
        )

    return return_value



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System

from duHast.Utilities.Objects.result import Result
from duHast.Revit.ColourFillSchemes.colour_fill_scheme_entry import get_entry_value_as_string, set_entry_value
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import ColorFillScheme, ColorFillSchemeEntry,Color,  ElementId, Transaction, StorageType

def update_by_values(doc, colour_fill_scheme, colour_fill_scheme_data, transaction_manager = in_transaction):
    """
    Updates the colours and fill patterns of a colour fill scheme by values.

    :param colour_fill_scheme: The colour fill scheme.
    :type colour_fill_scheme: ColourFillScheme
    :param colour_fill_scheme_data: The colour fill scheme data.
    :type colour_fill_scheme_data: dict
    
    :return: The result of the update operation.
    :rtype: Result
    """

    return_value = Result()

    try:
        # Check types

        if not isinstance(colour_fill_scheme, ColorFillScheme):
            raise TypeError("colour_fill_scheme must be of type ColourFillScheme. Got {}".format(type(colour_fill_scheme)))
        
        if not isinstance(colour_fill_scheme_data, list):
            raise TypeError("colour_fill_scheme_data must be of type list. Got {}".format(type(colour_fill_scheme_data)))


        def action():
            action_return_value = Result()

            try:
                # get all entries in the colour fill scheme
                entries = colour_fill_scheme.GetEntries()

                # loop over the entries and get the report data
                for entry in entries:

                    # get the value as string
                    parameter_value = get_entry_value_as_string(entry=entry)
                    
                    # set update flag to default
                    entry_requires_update = False

                    # loop over the colour fill scheme data and get the matching entry
                    for colour_fill_scheme_entry in colour_fill_scheme_data:
                        if parameter_value.lower() == colour_fill_scheme_entry.parameter_value.lower():
                            # create a color instance
                            new_colour = Color(
                                System.Convert.ToByte(colour_fill_scheme_entry.colour_red),
                                System.Convert.ToByte(colour_fill_scheme_entry.colour_green),
                                System.Convert.ToByte(colour_fill_scheme_entry.colour_blue),
                            )

                            # check if the colour is the same
                            if entry.Color.Red != new_colour.Red or entry.Color.Green != new_colour.Green or entry.Color.Blue != new_colour.Blue:
                                action_return_value.append_message("Colour is different: {} != {}".format(entry.Color, new_colour))
                                # update the entry with the new colour
                                entry.Color = new_colour
                                entry_requires_update = True


                            # check if the fill pattern is the same
                            if entry.FillPatternId.IntegerValue != colour_fill_scheme_entry.fill_pattern_id:
                                action_return_value.append_message("Fill pattern is different: {} != {}".format(entry.FillPatternId.IntegerValue, colour_fill_scheme_entry.fill_pattern_id))
                                # update the entry with the new fill pattern
                                entry.FillPatternId = ElementId(colour_fill_scheme_entry.fill_pattern_id)
                                entry_requires_update = True

                           
                            break

                    # this can throw an exception if the entry values are the same...
                    if entry_requires_update == False:
                        action_return_value.append_message("No update required for entry: {}".format(parameter_value))
                        continue
                    
                    # update the entry
                    colour_fill_scheme.UpdateEntry(entry)

                    # reset the update flag
                    entry_requires_update = False

                    # if the entry was updated, append the message
                    action_return_value.append_message("Updated entry: {} with colour: {},{},{} and fill pattern: {}".format(
                        parameter_value, 
                        colour_fill_scheme_entry.colour_red, 
                        colour_fill_scheme_entry.colour_green, 
                        colour_fill_scheme_entry.colour_blue, 
                        colour_fill_scheme_entry.fill_pattern_id)
                        )

                action_return_value.append_message("Colour fill scheme entries updated successfully.")
                return action_return_value
            except Exception as e:
                action_return_value.update_sep(False, "Error in action: {}".format(e))
                return action_return_value
        
        if transaction_manager is None:
           action_result =  action()
           return action_result
        else:
            tranny = Transaction(doc, "Updating colour fill scheme")
            result_delete = transaction_manager(tranny, action)
            return_value.update(result_delete)
            return return_value
       
    except Exception as e:
        return_value.update_sep(False, "Error updating colour fill scheme: {}".format(e))
        return return_value
   

def update_existing_and_add_new_values (doc, colour_fill_scheme, colour_fill_scheme_data, transaction_manager = in_transaction):
    """
    Updates the colours and fill patterns of a colour fill scheme by values and adds new values if provided in colour fill scheme data.

    :param colour_fill_scheme: The colour fill scheme.
    :type colour_fill_scheme: ColourFillScheme
    :param colour_fill_scheme_data: The colour fill scheme data.
    :type colour_fill_scheme_data: dict
    
    :return: The result of the update operation.
    :rtype: Result
    """

    return_value = Result()

    try:
        # Check types
        if not isinstance(colour_fill_scheme, ColorFillScheme):
            raise TypeError("colour_fill_scheme must be of type ColourFillScheme. Got {}".format(type(colour_fill_scheme)))
        
        if not isinstance(colour_fill_scheme_data, list):
            raise TypeError("colour_fill_scheme_data must be of type list. Got {}".format(type(colour_fill_scheme_data)))

        # define action to run inside a transaction
        def action():
            action_return_value = Result()

            try:
                # update existing values first
                update_existing_values_result = update_by_values(doc, colour_fill_scheme, colour_fill_scheme_data, transaction_manager = None)
                action_return_value.update(update_existing_values_result)

                # add new values
                # get all entries in the colour fill scheme
                entries = colour_fill_scheme.GetEntries()

                # loop over the colour fill scheme data and get the matching entry
                for colour_fill_scheme_entry in colour_fill_scheme_data:

                    entry_exists = False
                    # loop over the entries and get the report data
                    for entry in entries:

                        # get the value
                        parameter_value = get_entry_value_as_string(entry=entry)
                       
                        # check if the entry is already in the colour fill scheme
                        if parameter_value.lower() == colour_fill_scheme_entry.parameter_value.lower():
                            # skip this value as already updated in first step
                            entry_exists = True
                            break

                    # if the entry is in the colour fill scheme, skip it
                    if entry_exists == True:
                        entry_exists = False
                        continue

                    # create a new entry
                    new_entry = ColorFillSchemeEntry(StorageType(colour_fill_scheme_entry.storage_type))
                    
                    # set the value
                    new_entry = set_entry_value(new_entry, colour_fill_scheme_entry.parameter_value)

                    
                    # create a color instance
                    new_colour = Color(
                        System.Convert.ToByte(colour_fill_scheme_entry.colour_red),
                        System.Convert.ToByte(colour_fill_scheme_entry.colour_green),
                        System.Convert.ToByte(colour_fill_scheme_entry.colour_blue),
                    )
                   
                    # set the fill color
                    new_entry.Color = new_colour
                    
                    # set the fill pattern id
                    new_entry.FillPatternId = ElementId(colour_fill_scheme_entry.fill_pattern_id)
                    

                    # add the entry to the colour fill scheme
                    colour_fill_scheme.AddEntry(new_entry)
                    
                    action_return_value.append_message("Added new entry: {} with colour: {},{},{} and fill pattern: {}".format(
                        colour_fill_scheme_entry.parameter_value, 
                        colour_fill_scheme_entry.colour_red, 
                        colour_fill_scheme_entry.colour_green, 
                        colour_fill_scheme_entry.colour_blue, 
                        colour_fill_scheme_entry.fill_pattern_id)
                    )

                action_return_value.append_message("Colour fill scheme entries updated successfully.")
                return action_return_value
            except Exception as e:
                action_return_value.update_sep(False, "Error in action: {}".format(e))
                return action_return_value
            
        if transaction_manager is None:
           action_result =  action()
           return action_result
        else:
            tranny = Transaction(doc, "Updating colour fill scheme")
            result_delete = transaction_manager(tranny, action)
            return_value.update(result_delete)
            return return_value
       
    except Exception as e:
        return_value.update_sep(False, "Error updating colour fill scheme: {}".format(e))
        return return_value
   


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit Geometry to data geometry conversion helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Data.Objects.Collectors.Properties.Geometry import geometry_polygon_2 as dGeometryPoly
from duHast.Geometry.bounding_box_2 import BoundingBox2
from duHast.Geometry.point_2 import Point2
from duHast.Revit.Common.Geometry import geometry as rGeo, solids as rSolid
from duHast.Revit.Common.Geometry.points import get_point_as_doubles
from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm

from Autodesk.Revit.DB import XYZ


def convert_xyz_in_data_geometry_polygons(doc, dgObject):
    """
    Converts Revit XYZ objects stored in a data geometry object into groups of doubles for inner and outer loops\
        and stores them in new data geometry object. It also populates translation and rotation matrix data of\
            coordinate system information.
    :param doc: _description_
    :type doc: _type_
    :param dgObject: A data geometry object.
    :type dgObject: :class:`.DataGeometryPolygon`
    :return: A data geometry object.
    :rtype: :class:`.DataGeometryPolygon`
    """

    data_geometry = dGeometryPoly.DataGeometryPolygon2()
    outer_loop = []
    for xyz_point in dgObject.outer_loop:
        point_as_double = get_point_as_doubles(xyz_point)
        outer_loop.append(point_as_double)
    inner_loops = []
    for inner_loop in dgObject.inner_loops:
        inner_loop_points = []
        for xyz_point in inner_loop:
            point_as_double = get_point_as_doubles(xyz_point)
            inner_loop_points.append(point_as_double)
        inner_loops.append(inner_loop_points)
    data_geometry.outer_loop = outer_loop
    data_geometry.inner_loops = inner_loops
    # add coordinate system translation and rotation data
    (
        data_geometry.rotation_coord,
        data_geometry.translation_coord,
    ) = rGeo.get_coordinate_system_translation_and_rotation(doc)
    return data_geometry


def convert_bounding_box_to_flattened_2d_points(bounding_box):
    """
    Converts a bounding box into a 2D polygon by projecting it onto a plane.( Removes Z values...)
    :param bounding_box: A bounding box.
    :type bounding_box: Autodesk.Revit.DB.BoundingBoxXYZ
    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometryPolygon`
    """
    # get points from bounding box and project them onto a plane ( Z value of bounding box min point)
    bounding_box_points = [
        bounding_box.Min,
        XYZ(bounding_box.Max.X, bounding_box.Min.Y, bounding_box.Min.Z),
        XYZ(bounding_box.Max.X, bounding_box.Max.Y, bounding_box.Min.Z),
        XYZ(bounding_box.Min.X, bounding_box.Max.Y, bounding_box.Min.Z),
    ]
    # set up data class object and store points in outer loop property
    dataGeometry = dGeometryPoly.DataGeometryPolygon2()
    dataGeometry.outer_loop = bounding_box_points
    return dataGeometry


def get_2d_points_from_revit_element_type_in_model(doc, element_instance_getter):
    """
    Returns a list of lists of points representing the flattened(2D geometry) of the elements
    List of Lists because a elements can be made up of multiple solids. Each nested list represents one solid within the elements geometry.
    Does not work with in place elements.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_instance_getter: Function returning all element instances of a particular category in the model as an element collector
    :type element_instance_getter: func(doc)

    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometry`
    """

    element_instances = element_instance_getter(doc)
    all_element_points = []
    for element_instance in element_instances:
        element_points = rSolid.get_2d_points_from_solid(element_instance)
        if len(element_points) > 0:
            all_element_points.append(element_points)
    return all_element_points


def convert_revit_bounding_box_to_geometry2_bounding_box(bounding_box):
    
    point1 = Point2(
        x=convert_imperial_feet_to_metric_mm(bounding_box.Min.X),
        y=convert_imperial_feet_to_metric_mm(bounding_box.Min.Y),
    )
    point2 = Point2(
        x=convert_imperial_feet_to_metric_mm(bounding_box.Max.X),
        y=convert_imperial_feet_to_metric_mm(bounding_box.Max.Y),
    )
    bbox = BoundingBox2(point1=point1, point2=point2)
    return bbox



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family report data utility module containing functions to manipulate report files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.Data.family_report_reader import read_data_into_families
from duHast.Utilities.files_io import file_exist


def _check_families_still_exist(family_data):
    """
    Checks whether families still exist on file server.

    Reason why families no longer exist:

    - family got deleted or moved
    - family got renamed

    :param famData: A list containing FamilyDataFamily instances.
    :type famData: [:class:`.FamilyDataFamily`]

    :return:
        Result class instance.

        - .status True if successfully removed any outdated family data or None needed removing. Otherwise False.
        - .message will contain list of families removed or message nothing needed to be removed.
        - . result will contain new list of FamilyDataFamily instances which still exist on the file server

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = Result()
    filtered_list = []
    family_was_removed = False
    try:
        # check which ones do not exist anymore
        for family_data_instance in family_data:
            if file_exist(family_data_instance.family_file_path):
                filtered_list.append(family_data_instance)
            else:
                return_value.append_message(
                    "Removed family: {}".format(family_data_instance.family_file_path)
                )
                family_was_removed = True

        if family_was_removed == False:
            return_value.append_message("No family required removing from data.")

        # update return data
        return_value.update_sep(True, "Successfully updated family data.")
        return_value.result = filtered_list

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to check whether families still exist with exception: {}".format(e),
        )
    return return_value


def combine_reports(previous_report_path, new_report_path):
    """
    This combines family reports:

    - reading reports into FamilyDataFamily instances
    - comparing the previous report families with the new report families and
        - adding any new families found in the new reports
        - adding previous report families not found in new report families

    This function checks at the end whether families still exist on file server. If not, they will be removed from the data set.

    :param previous_report_path: A fully qualified file path to the previous report file.
    :type previous_report_path: str
    :param new_report_path: A fully qualified file path to the new report file.
    :type new_report_path: str

    :return:
        Result class instance.

        - .status True if successfully combined report(s). Otherwise False.
        - .message will contain count of new families added and previous families retained.
        - . result will contain list of FamilyDataFamily instances representing the combined data set

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = Result()
    combined_families = []
    try:
        # checks:
        # are past in path strings
        if isinstance(previous_report_path, str) == False:
            raise TypeError(
                "previous_report_path should be of type string but is: {}".format(
                    type(previous_report_path)
                )
            )
        if isinstance(new_report_path, str) == False:
            raise TypeError(
                "new_report_path should be of type string but is: {}".format(
                    type(new_report_path)
                )
            )

        # read values from previous path
        previous_families_read_result = read_data_into_families(
            path_to_data=previous_report_path
        )
        # check read was successful
        if previous_families_read_result.status == False:
            raise ValueError(previous_families_read_result.message)
        else:
            return_value.append_message(
                "Successfully read {} families from: {}".format(
                    len(previous_families_read_result.result), previous_report_path
                )
            )

        # read values from new path
        new_families_read_result = read_data_into_families(path_to_data=new_report_path)
        # check read was successful
        if new_families_read_result.status == False:
            raise ValueError(new_families_read_result.message)
        else:
            return_value.append_message(
                "Successfully read {} families from: {}".format(
                    len(new_families_read_result.result), new_report_path
                )
            )

        # some stats:
        new_families_added = 0
        previous_families_retained = 0

        # check if any new families where retrieved
        if len(new_families_read_result.result) > 0:
            new_families_added = len(new_families_read_result.result)
            combined_families = new_families_read_result.result
            # take new families as base line and append previous report families only which have no match in new families
            for previous_family in previous_families_read_result.result:
                if previous_family not in combined_families:
                    combined_families.append(previous_family)
                    previous_families_retained += 1
        else:
            # return previous families unchanged since no new families exist
            combined_families = previous_families_read_result.result
            previous_families_retained = len(previous_families_read_result.result)

        # set the return value
        return_value.append_message(
            "Combined report contains {} new families and {} previous report families where retained.".format(
                new_families_added, previous_families_retained
            )
        )
        return_value.result = combined_families

        # check if all families still exist on the server...if not remove from list
        check_files_exists = _check_families_still_exist(combined_families)
        if check_files_exists.status:
            if len(check_files_exists.result) != len(combined_families):
                return_value.append_message(
                    "Removed {} families from data set as they no longer exist on server.".format(
                        len(combined_families) - len(check_files_exists.result)
                    )
                )
                return_value.result = check_files_exists.result
            else:
                return_value.append_message(
                    "All families still exist on server. None was removed from data set."
                )
        else:
            return_value.update_sep(
                False,
                "Failed to check file exists with exception: {}".format(
                    check_files_exists.message
                ),
            )
            # reset the result to an empty list!
            return_value.result = []

    except Exception as e:
        return_value.update_sep(
            False, "An exception ocurred when combining reports: {}".format(e)
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit category helper functions for project files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
from System import Enum
from collections import namedtuple

from Autodesk.Revit.DB import BuiltInCategory, Category


# tuples containing categories data
category_data = namedtuple("category_data", "category_name sub_category_name id")


def get_categories_in_model(doc):
    """
    Returns all categories and subcategories in a model

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of named tuples of type category_data
    :rtype: [category_data]
    """

    cats = doc.Settings.Categories
    categories = []
    for main_category in cats:
        categories.append(
            category_data(
                category_name=main_category.Name,
                sub_category_name=main_category.Name,
                id=main_category.Id,
            )
        )
        for sub_cat in main_category.SubCategories:
            categories.append(
                category_data(
                    category_name=main_category.Name,
                    sub_category_name=sub_cat.Name,
                    id=sub_cat.Id,
                )
            )

    return categories


def get_main_categories_in_model(doc):
    """
    Returns all main categories in a model
    
    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    
    :return: List of named tuples of type category_data
    :rtype: [category_data]
    """
    
    cats = doc.Settings.Categories
    categories = []
    for main_category in cats:
        categories.append(
            category_data(
                category_name=main_category.Name,
                sub_category_name=main_category.Name,
                id=main_category.Id,
            )
        )
    return categories


def get_category_from_builtInCategory(doc, built_in_category):
    """
    Returns a category based on the build in category enum value.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param built_in_category: The built in category
    :type built_in_category: Autodesk.Revit.DB.BuiltInCategory
    :return: A category
    :rtype: Autodesk.Revit.DB.Category
    """

    return Category.GetCategory(doc, built_in_category)


def get_builtInCategory_from_category(doc, category):
    """
    Returns a built in category enum value based on the category object.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param category: The category object
    :type category: Autodesk.Revit.DB.Category
    :return: The built in category enum value
    :rtype: Autodesk.Revit.DB.BuiltInCategory
    """

    values = Enum.GetValues(BuiltInCategory)
    names = Enum.GetNames(BuiltInCategory)
    for value, name in zip(values, names):
        try:
            cat = Category.GetCategory(doc, value)
            if cat == None:
                continue
            if cat.Name == category.Name:
                return value
        except:
            continue
    return None


def get_category_by_names(doc, main_category_name, sub_category_name):
    """
    Retrieves a category object from a document based on the given main category name and optional sub category name.

    Args:
        doc (document object): The document object from which to retrieve the category.
        main_category_name (str): The name of the main category to search for.
        sub_category_name (str): The name of the sub category to search for within the main category.

    Returns:
        category object: The retrieved category object if a match is found.
        None: If no match is found.
    """
    return_value = None
    cats = doc.Settings.Categories
    for main_category in cats:
        if main_category.Name == main_category_name:
            if sub_category_name == main_category_name:
                return main_category
            else:
                for sub_cat in main_category.SubCategories:
                    if sub_cat.Name == sub_category_name:
                        return sub_cat
    return return_value


def get_builtin_category_by_name(category_name):
    """
    Returns the built in category enum value based on the category name.(i.e. OST_Walls)

    :param category_name: The category name
    :type category_name: str
    :return: The built in category enum value
    :rtype: Autodesk.Revit.DB.BuiltInCategory
    """

    values = Enum.GetValues(BuiltInCategory)
    names = Enum.GetNames(BuiltInCategory)
    for value, name in zip(values, names):
        if name == category_name:
            return value
    return None



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A logger class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The LoggerObject class is a custom logger class that extends both logging.Logger and Base.
It is designed to handle logging with both file and console handlers, providing a structured
and formatted logging mechanism.

Key Features:
1. Initialization:
   - The constructor initializes the logger with a specified name, output path, log levels,
     file format, and handlers for both file and console logging.
   - It creates the necessary directories and sets up the log file path.
   - Handlers are set up with specified formatters and added to the logger.

2. Methods:
   - clear_handlers(): Clears all existing handlers from the logger.
   - init_handlers(): Initializes and adds the file and console handlers to the logger.
   - get_logger_obj(): Returns the logger object.
   - update_log_level(log_level): Updates the log levels for both file and console handlers.

Usage:
- The class can be used to create a logger that writes logs to a file and outputs them to the console
  with specified formats.
- It allows dynamic updating of log levels and ensures that handlers are properly managed.

Example Usage:

# Create a logger object
logger = LoggerObject(log_name="myLogger", output_path="/path/to/logs")

# Get the logger object
logger_obj = logger.get_logger_obj()

# Update log levels
logger.update_log_level((logging.INFO, logging.ERROR))

# Log messages
logger_obj.info("This is an info message")
logger_obj.error("This is an error message")


"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import logging
from logging.handlers import RotatingFileHandler
import os

from duHast.Utilities.directory_io import create_target_directory
from duHast.Utilities.Objects.base import Base
from duHast.Utilities.logging_handlers import std_console_handler, std_file_handler
from duHast.Utilities.logger_formatting import level_time_local_msg_formatter, msg_only


class LoggerObject(logging.Logger, Base):
    def __init__(
        self,
        log_name="duHast",
        output_path=os.getenv("APPDATA"),
        log_level=(10, 30),
        file_format=".txt",
        fil_stream_hndlr=RotatingFileHandler,
        fil_stream_frmt=level_time_local_msg_formatter,
        cons_stream_hndlr=logging.StreamHandler,
        cons_stream_frmt=msg_only,
        **kwargs
    ):
        """
        Constructor for the LoggerObject class.

        :param log_name: The name of the log.
        :type log_name: str
        :param output_path: The path to the output directory.
        :type output_path: str
        :param log_level: The log levels for the file and console handlers.
        :type log_level: tuple
        :param file_format: The file format for the log file.
        :type file_format: str
        :param fil_stream_hndlr: The file stream handler class.
        :type fil_stream_hndlr: class
        :param fil_stream_frmt: The file stream formatter.
        :type fil_stream_frmt: logging.Formatter
        :param cons_stream_hndlr: The console stream handler class.
        :type cons_stream_hndlr: class
        :param cons_stream_frmt: The console stream formatter.
        :type cons_stream_frmt: logging.Formatter
        """
        
        # Get log naming and output path
        self.log_name = log_name
        self.output_path = output_path
        self.log_file_dir = os.path.join(self.output_path, self.log_name)
        # Create the directory and the full log file path
        create_target_directory(self.output_path, self.log_name)
        self.log_file_path = os.path.join(self.log_file_dir, log_name + file_format)
        logging.Logger.__init__(self, log_name)
        logging.basicConfig(filename=self.log_file_path, level=logging.INFO)
        # Get log levels
        self.level = log_level
        self.file_log_level, self.console_log_level = self.level
        # Set handlers
        self.file_handler = std_file_handler(
            self, file_hndlr=fil_stream_hndlr, file_formatter=fil_stream_frmt
        )
        self.console_handler = std_console_handler(
            self, cons_hndlr=cons_stream_hndlr, cons_formatter=cons_stream_frmt
        )

        # Create logger object
        self.new_logger = logging.getLogger(self.log_name)
        self.init_handlers()

    def clear_handlers(self):
        """
        Clears all handlers from the logger.
        """
        
        self.new_logger.propagate = 0
        for handler in self.new_logger.handlers:
            handler.flush()
            handler.close()
            self.new_logger.removeHandler(handler)

    def init_handlers(self):
        """
        Initializes the file and console handlers for the logger.
        """
        
        self.clear_handlers()
        ex_handlers = [h.__class__.__name__ for h in self.new_logger.handlers]
        self.new_logger.setLevel(logging.INFO)

        if not "FileHandler" in ex_handlers:
            self.new_logger.addHandler(self.file_handler)
        if not "StreamHandler" in ex_handlers:
            self.new_logger.addHandler(self.console_handler)

    def get_logger_obj(self):
        """
        Returns the logger object.

        :return: The logger object.
        :rtype: logging.Logger
        """
        
        return self.new_logger

    def update_log_level(self, log_level):
        """
        Set the log levels for the file and console handlers.

        :param log_level: A tuple containing the log levels for the file and console handlers.
        :type log_level: tuple
        :return: The updated logger object.
        :rtype: logging.Logger
        """
        
        log_level_file, log_level_console = log_level
        self.file_handler.setLevel(log_level_file)
        self.console_handler.setLevel(log_level_console)
        return self.get_logger_obj()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System


clr.AddReference("System.Core")
clr.ImportExtensions(System.Linq)
clr.AddReference("System")


# import common library
# utility functions for most commonly used Revit API tasks
from duHast.Revit.Common import parameter_get_utils as rParaGet

# utilities
from duHast.Utilities import files_io as fileIO

# class used for stats reporting
from duHast.Utilities.Objects import result as res

# implementation of Revit API callback required when loading families into a Revit model
from duHast.Revit.Family import family_load_option as famLoadOpt

# load everything required from family load call back
from duHast.Revit.Family.family_load_option import *
from duHast.Revit.Common import transaction as rTran
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Revit.Common.failure_handling import process_failures
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)

from duHast.Revit.Common.parameter_get_utils import (
    get_built_in_parameter_value,
    get_parameter_value_as_integer,
)

# import Autodesk Revit DataBase namespace
from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    ElementCategoryFilter,
    ElementClassFilter,
    ElementMulticategoryFilter,
    ElementId,
    ElementParameterFilter,
    Family,
    FamilyInstance,
    FamilySymbol,
    FilteredElementCollector,
    FilterElementIdRule,
    FilterNumericEquals,
    ParameterValueProvider,
    Transaction,
)

from duHast.Revit.Family.Utility.loadable_family_categories import (
    CATEGORIES_LOADABLE_3D,
    CATEGORIES_LOADABLE_TAGS,
)

# --------------------------------------------------- Family Loading / inserting -----------------------------------------

def load_family(doc, family_file_path, transaction_manager=in_transaction_with_failure_handling, failure_config = FailureHandlingConfig(), failure_processing_func=process_failures):
    """
    Loads or reloads a single family into a Revit document.

    Will load/ reload family provided in in path. By default the parameter values in the project file will be overwritten
    with parameter values in family.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_file_path: The fully qualified file path of the family to be loaded.
    :type family_file_path: str
    :param  transaction_manager: A function wrapping the load family action in a transaction. Default is in_transaction_with_failure_handling. None assumes a transaction is already in progress.
    :type transaction_manager: function
    :param failure_config: The failure handling configuration object. Default is FailureHandlingConfig()
    :type failure_config: FailureHandlingConfig
    :param failure_processing_func: The function to process failures. Default is process_failures
    :type failure_processing_func: function
    
    :raise: None

    :return:
        Result class instance.

        - Reload status (bool) returned in result.status.
        - Reload message: contains the reload log messages.
        - Return family reference stored in result.result property on successful reload only

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        # set up load / reload action to be run within a transaction
        def action():
            # set up return value for the load / reload
            return_family = clr.Reference[Family]()
            action_return_value = res.Result()
            try:
                # attempt to reload the family
                reload_status = doc.LoadFamily(
                    family_file_path,
                    famLoadOpt.FamilyLoadOption(),  # overwrite parameter values etc
                    return_family,
                )
                # update reload status
                action_return_value.update_sep(
                    reload_status,
                    "Loaded family: {} :: {}".format(family_file_path, reload_status),
                )
                
                # check if reload returned a family and if so return it to caller
                # needs to be 'is not None test since an exception will thrown otherwise:
                # Can't convert a Reference<> instance to a bool
                if return_family is not None:
                    action_return_value.result.append(return_family.Value)
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to load family: {} with exception: {}".format(
                        family_file_path, e
                    ),
                )
            return action_return_value

        # if no transaction manager is passed in, assume a transaction is already in progress
        if transaction_manager is None:
            # load family without starting a new transaction
            return_value.update(action())
        else:
            transaction = Transaction(
                doc,
                "Loading Family: {}".format(
                    fileIO.get_file_name_without_ext(family_file_path)
                ),
            )
            dummy = transaction_manager(transaction, action, failure_config, failure_processing_func)
            return_value.update(dummy)
    except Exception as e:
        return_value.update_sep(False, "Failed to load family with exception: {}".format(e))
    return return_value



# ------------------------ filter functions -------------------------------------------------------------------------------------


def get_family_symbols(doc, cats):
    """
    Filters all family symbols (Revit family types) of given built in categories from the Revit model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param cats: ICollection of Autodesk.Revit.DB.BuiltInCategory values.
    :type cats: ICollection
        :cats sample: cats = List[rdb.BuiltInCategory] ([rdb.BuiltInCategory.OST_Furniture, rdb.BuiltInCategory.OST_Parking])

    :return: A collector of Autodesk.Revit.DB.Element matching filter.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    elements = []
    try:
        multi_cat_filter = ElementMulticategoryFilter(cats)
        elements = (
            FilteredElementCollector(doc)
            .OfClass(FamilySymbol)
            .WherePasses(multi_cat_filter)
            .ToElements()
        )
        return elements
    except Exception:
        return elements


def get_family_instances_by_built_in_categories(doc, cats):
    """
    Filters all family instances of given built in categories from the Revit model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param cats: ICollection of Autodesk.Revit.DB.BuiltInCategory values:
    :type cats: ICollection
        :cats sample: cats = List[rdb.BuiltInCategory] ([rdb.BuiltInCategory.OST_Furniture, rdb.BuiltInCategory.OST_Parking])

    :return: A collector of Autodesk.Revit.DB.Element matching filter.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    elements = []
    try:
        multi_cat_filter = ElementMulticategoryFilter(cats)
        elements = (
            FilteredElementCollector(doc)
            .OfClass(FamilyInstance)
            .WherePasses(multi_cat_filter)
            .ToElements()
        )
        return elements
    except Exception:
        return elements


def get_family_instances_of_built_in_category(doc, builtin_cat):
    """
    Filters all family instances of a single given built in category from the Revit model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param builtin_cat: single revit builtInCategory Enum value.
    :type builtin_cat: Autodesk.Revit.DB.BuiltInCategory

    :return: A collector of Autodesk.Revit.DB.FamilyInstance matching filter.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = ElementCategoryFilter(builtin_cat)
    col = (
        FilteredElementCollector(doc)
        .OfClass(FamilyInstance)
        .WherePasses(filter)
    )
    return col


def get_all_loadable_families(doc):
    """
    Filters all families in revit model by whether it is not an InPlace family.

    Note: slow filter due to use of lambda and cast to list.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of families matching filter.
    :rtype: list Autodesk.Revit.DB.Family
    """

    collector = FilteredElementCollector(doc)
    families = (
        collector.OfClass(Family).Where(lambda e: (e.IsInPlace == False)).ToList()
    )
    return families


def get_all_loadable_family_ids_through_types(doc):
    """
    Get all loadable family ids in file.

    :param doc: Current family document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of family ids
    :rtype: [Autodesk.Revit.DB.ElementId]
    """

    family_ids = []
    col = FilteredElementCollector(doc).OfClass(FamilySymbol)
    # get families from symbols and filter out in place families
    for fam_symbol in col:
        if (
            fam_symbol.Family.Id not in family_ids
            and fam_symbol.Family.IsInPlace == False
        ):
            family_ids.append(fam_symbol.Family.Id)
    return family_ids


def get_all_in_place_families(doc):
    """
    Filters all families in revit model by whether it is an InPlace family.

    Note: slow filter due to use of lambda and cast to list

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of families matching filter.
    :rtype: list Autodesk.Revit.DB.Family
    """

    collector = FilteredElementCollector(doc)
    families = (
        collector.OfClass(Family).Where(lambda e: (e.IsInPlace == True)).ToList()
    )
    return families


def get_all_family_instances(doc):
    """
    Returns all family instances in document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A collector with all family instances in document.
    :rtype: Autodesk.Revit.DB.Collector
    """

    col = FilteredElementCollector(doc).OfClass(FamilyInstance)
    return col


# --------------------------family data ----------------


def is_any_nested_family_instance_label_driven(doc):
    """
    Checks whether any family instance in document is driven by the 'Label' property.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: True if at least one instance is driven by label property. Otherwise False
    :rtype: bool
    """

    flag = False
    fam_instances = get_all_family_instances(doc)

    for fam_instance in fam_instances:
        # get the Label parameter value
        p_value = rParaGet.get_built_in_parameter_value(
            fam_instance,
            BuiltInParameter.ELEM_TYPE_LABEL,
            rParaGet.get_parameter_value_as_element_id,
        )
        # a valid Element Id means family instance is driven by Label
        if p_value != ElementId.InvalidElementId:
            flag = True
            break

    return flag


def get_symbols_from_type(doc, type_ids):
    """
    Get all family types belonging to the same family as types past in.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_ids: - list of element id's representing family symbols (family types)
    :type type_ids: list of Autodesk.Revit.DB.ElementId

    :return: dictionary:
        where key is the family id as Autodesk.Revit.DB.ElementId
        value is a list of all symbol(family type) ids as Autodesk.Revit.DB.ElementId belonging to the family
    :rtype: dic {Autodesk.Revit.DB.ElementId: list[Autodesk.Revit.DB.ElementId]}
    """

    families = {}
    for t_id in type_ids:
        # get family element
        type_el = doc.GetElement(t_id)
        fam_el = type_el.Family
        # check whether family was already processed
        if fam_el.Id not in families:
            # get all available family types
            s_ids = fam_el.GetFamilySymbolIds().ToList()
            families[fam_el.Id] = s_ids
    return families


def get_family_instances_by_symbol_type_id(doc, type_id):
    """
    Filters all family instances of a single given family symbol (type).

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param Autodesk.Revit.DB.ElementId type_id: The symbol (type) id

    :return: A collector of Autodesk.Revit.DB.FamilyInstance matching filter.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    pvp_symbol = ParameterValueProvider(
        ElementId(BuiltInParameter.SYMBOL_ID_PARAM)
    )
    equals = FilterNumericEquals()
    id_filter = FilterElementIdRule(pvp_symbol, equals, type_id)
    element_filter = ElementParameterFilter(id_filter)
    collector = FilteredElementCollector(doc).WherePasses(element_filter)
    return collector


def get_family_instances_by_symbol_type_fast(family_symbol):
    """
    Return number of instances in the model ( fast!)

    :param family_symbol: A family type
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol

    :return: A c# IList of dependent element ids.
    :rtype: IList<ElementId>
    """

    filter = ElementClassFilter(FamilyInstance)
    dependent_element_ids = family_symbol.GetDependentElements(filter)

    # this is an IList
    return dependent_element_ids


def get_all_in_place_type_ids_in_model_of_category(doc, fam_built_in_category):
    """
    Filters family symbol (type) ids off all available in place families of single given built in category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param fam_built_in_category: built in revit category
    :type fam_built_in_category: Autodesk.Revit.DB.BuiltInCategory

    :return: A list of Element Ids representing the family symbols matching filter.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    # filter model for family symbols of given built in category
    filter = ElementCategoryFilter(fam_built_in_category)
    col = (
        FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter)
    )
    ids = []
    for c in col:
        fam = c.Family
        # check if this an in place or loaded family!
        if fam.IsInPlace == True:
            ids.append(c.Id)
    return ids


# --------------------------family purge  ----------------


def get_family_symbols_ids(doc, cats, exclude_shared_fam=True):
    """
    Filters family symbols belonging to list of built in categories past in.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ICollection cats: ICollection of Autodesk.Revit.DB.BuiltInCategory values.
    :type cats: cats = List[rdb.BuiltInCategory] ([rdb.BuiltInCategory.OST_Furniture, rdb.BuiltInCategory.OST_Parking])

    :return: A list of Element Ids representing the family symbols matching filter.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    try:
        multi_cat_filter = ElementMulticategoryFilter(cats)
        elements = (
            FilteredElementCollector(doc)
            .OfClass(FamilySymbol)
            .WherePasses(multi_cat_filter)
        )
        for el in elements:
            # check if shared families are to be excluded from return list
            if exclude_shared_fam:
                fam = el.Family
                p_value = rParaGet.get_built_in_parameter_value(
                    fam, BuiltInParameter.FAMILY_SHARED
                )
                if p_value != None:
                    if p_value == "No" and el.Id not in ids:
                        ids.append(el.Id)
                else:
                    # some revit families cant be of type shared...()
                    ids.append(el.Id)
            else:
                ids.append(el.Id)
        return ids
    except Exception:
        return ids


def get_all_non_shared_family_symbol_ids(doc):
    """
    Filters family symbols (types) belonging to hard coded categories lists (catsLoadableThreeD, catsLoadableTags)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of Element Ids representing the family symbols matching filter.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    all_loadable_three_d_type_ids = get_family_symbols_ids(doc, CATEGORIES_LOADABLE_3D)
    all_loadable_tags_type_ids = get_family_symbols_ids(doc, CATEGORIES_LOADABLE_TAGS)
    ids = all_loadable_three_d_type_ids + all_loadable_tags_type_ids
    return ids


def is_shared_from_family(family):
    """
    Returns if the family is shared (true) or not (false)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_symbol: The family 
    :type family_symbol: Autodesk.Revit.DB.Family

    :return: True if family is shared, otherwise False
    :rtype: bool
    """

    p_value = get_built_in_parameter_value(
        element=family,
        built_in_parameter_def=BuiltInParameter.FAMILY_SHARED,
        parameter_value_getter=get_parameter_value_as_integer,
    )
    
    if p_value==0:
        return False
    else:
        return True
    
    
def is_shared_from_symbol(family_symbol):
    """
    Returns if the family is shared (true) or not (false)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_symbol: The family symbol (type)
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol

    :return: True if family is shared, otherwise False
    :rtype: bool
    """

    fam = family_symbol.Family
    return is_shared_from_family(family=fam)
    

    


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to load or reload families 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

clr.AddReference("System")
from System.Collections.Generic import List


# import common library
from duHast.Revit.Common import delete as rDel

from duHast.Utilities import files_get as fileGet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family import family_utils as rFamUtil
from duHast.Revit.Family import family_load_option as famLoadOpt
from duHast.Revit.Family.family_load_option import *
from duHast.Revit.Family.Utility import loadable_family_categories as rFamLoadable

import Autodesk.Revit.DB as rdb

# --------------------------------------------------- Family Loading / inserting -----------------------------------------


def reload_all_families(doc, library_location, include_sub_folders):
    """
    Reloads a number of families with setting: parameter values overwritten: True


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param library_location: _description_
    :type library_location: str
    :param include_sub_folders: _description_
    :type include_sub_folders: bool

    :raises UserWarning: _description_

    :return: Returns True if any of the reload actions was successful.
    :rtype: bool
    """

    result = res.Result()
    # if a family is reloaded it may bring in new types not present in the model at reload
    # this list contains the ids of those types (symbols)
    # so they can be deleted if so desired
    symbol_ids_to_be_deleted = []
    try:
        # build library
        library = fileGet.files_as_dictionary(
            library_location, "", "", ".rfa", include_sub_folders
        )
        if len(library) == 0:
            result.update_sep(False, "Library is empty!")
            # get out...
            raise UserWarning("Empty Library")
        else:
            result.append_message("Found: {} families in Library!".format(len(library)))
        # get all families in file:
        family_ids = get_family_ids_from_symbols(doc)
        if len(family_ids) > 0:
            result.append_message(
                "Found:  {} loadable families in file.".format(len(family_ids))
            )
            for fam_id in family_ids:
                fam = doc.GetElement(fam_id)
                fam_name = rdb.Element.Name.GetValue(fam)
                if fam_name in library:
                    result.append_message("Found match for: {}".format(fam_name))
                    if len(library[fam_name]) == 1:
                        # found single match for family by name
                        result.update_sep(
                            True, "Found single match: {}".format(library[fam_name][0])
                        )
                        # get all symbols attached to this family by name
                        prior_load_symbol_ids = fam.GetFamilySymbolIds()
                        # reload family
                        result_load = rFamUtil.load_family(doc, library[fam_name][0])
                        result.append_message(result_load.message)
                        if result_load.status == True:
                            # make sure that if a single reload was successful that this method returns true
                            result.status = True
                            # remove symbols (family types) added through reload process
                            if (
                                result_load.result != None
                                and len(result_load.result) > 0
                            ):
                                fam_loaded = result_load.result[0]
                                after_load_symbol_ids = fam_loaded.GetFamilySymbolIds()
                                new_symbol_ids = get_new_symbol_ids(
                                    prior_load_symbol_ids, after_load_symbol_ids
                                )
                                if len(new_symbol_ids) > 0:
                                    symbol_ids_to_be_deleted = (
                                        symbol_ids_to_be_deleted + new_symbol_ids
                                    )
                    else:
                        matches_message = ""
                        for path in library[fam_name]:
                            matches_message = matches_message + "..." + path + "\n"
                        matches_message = "Found multiple matches for {} \n {}".format(
                            fam_name, matches_message
                        )
                        matches_message = matches_message.strip()
                        # found multiple matches for family by name only...aborting reload
                        result.append_message(matches_message)
                else:
                    result.update_sep(
                        result.status, "Found no match for: {}".format(fam_name)
                    )
            # delete any new symbols introduced during the reload
            if len(symbol_ids_to_be_deleted) > 0:
                result_delete = rDel.delete_by_element_ids(
                    doc,
                    symbol_ids_to_be_deleted,
                    "Delete new family types",
                    "Family types",
                )
                result.append_message(result_delete.message)
            else:
                message = "No need to delete any new family types since no new types where created."
                result.append_message(message)  # make sure not to change the status
        else:
            message = "Found no loadable families in file!"
            result.update_sep(False, message)
    except Exception as e:
        message = "Failed to load families with exception: {}".format(e)
        result.update_sep(False, message)
    return result


def get_family_ids_from_symbols(doc):
    """
    Get all loadable family ids in file

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing loadable families.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    family_ids = []
    # build list of all categories we want families to be reloaded of
    fam_cats = List[rdb.BuiltInCategory](rFamLoadable.CATEGORIES_LOADABLE_TAGS)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_TAGS_OTHER)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_3D)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_3D_OTHER)
    # get all symbols in file
    fam_symbols = rFamUtil.get_family_symbols(doc, fam_cats)
    # get families from symbols and filter out in place families
    for fam_symbol in fam_symbols:
        if (
            fam_symbol.Family.Id not in family_ids
            and fam_symbol.Family.IsInPlace == False
        ):
            family_ids.append(fam_symbol.Family.Id)
    return family_ids


def get_new_symbol_ids(pre_load_symbol_id_list, after_load_symbol_list):
    """
    Returns a list of symbol ids not present prior to reload.

    Compares past in list of id's and returns ids not in preloadSymbolIdList

    :param pre_load_symbol_id_list: List of Ids of symbols prior the reload.
    :type pre_load_symbol_id_list: list of Autodesk.Revit.DB.ElementId
    :param after_load_symbol_list: List of ids of symbols after the reload.
    :type after_load_symbol_list: list of Autodesk.Revit.DB.ElementId

    :return: List of element ids representing Family Symbols.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    for id in after_load_symbol_list:
        if id not in pre_load_symbol_id_list:
            ids.append(id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Class for family type data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Used as accumulate all family type storage class instances when retrieving family type data from a family file.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.base import Base
from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)


class FamilyTypeDataStorageManager(Base):

    def __init__(self):

        super(FamilyTypeDataStorageManager, self).__init__()

        # storage for family type data
        self._family_type_data_storage = []

        # set up fields for family name and category
        self._family_name = None
        self._family_category = None
        self._family_file_path = None

    @property
    def family_type_data_storage(self):
        return self._family_type_data_storage

    @property
    def family_name(self):
        return self._family_name

    @family_name.setter
    def family_name(self, family_name):
        # only allow to set the family name if its current value is None
        if self._family_name is not None:
            raise ValueError("family_name can only be set once")
        self._family_name = family_name

    @property
    def family_category(self):
        return self._family_category

    @family_category.setter
    def family_category(self, family_category):
        # only allow to set the family category if its current value is None
        if self._family_category is not None:
            raise ValueError("family_category can only be set once")
        self._family_category = family_category


    @property
    def family_file_path(self):
        """
        Get the family file path

        :return: the family file path
        :rtype: str
        """
        return self._family_file_path
    
    @family_file_path.setter
    def family_file_path(self, family_file_path):
        """
        Set the family file path

        :param family_file_path: the family file path
        :type family_file_path: str
        """
        # only allow to set the family file path if its current value is None
        if self._family_file_path is not None:
            raise ValueError("family_file_path can only be set once")
        self._family_file_path = family_file_path


    @property
    def family_has_types(self):
        """
        Check if the family has types
        (might have not types if all types where purge from a model and just the family was left)

        :return: True if the family has types, False otherwise
        :rtype: bool
        """

        return len(self.family_type_data_storage) > 0


    def add_family_type_data_storage(self, family_type_data_storage):
        """
        add family type data storage to the manager.

        This will also set the family name and category of this storage manager instance if they are not set yet.

        :param family_type_data_storage: family type data storage instance
        :type family_type_data_storage: FamilyTypeDataStorage
        """

        if not isinstance(family_type_data_storage, FamilyTypeDataStorage):
            raise ValueError(
                "family_type_data_storage must be an instance of FamilyTypeDataStorage"
            )

        # check if the family name and category are set
        if self.family_name is None:
            self.family_name = family_type_data_storage.family_name
        if self.family_category is None:
            self.family_category = family_type_data_storage.root_category_path

        # check if the family name and category match the family type data storage instance
        if self.family_name != family_type_data_storage.family_name:
            raise ValueError(
                "family_name does not match the family type data storage instance"
            )
        if self.family_category != family_type_data_storage.root_category_path:
            raise ValueError(
                "family_category does not match the family type data storage instance"
            )

        self._family_type_data_storage.append(family_type_data_storage)


    def remove_family_type(self, family_type_name):
        """
        Remove a family type from the family type data storage manager

        :param family_type_name: the name of the family type to remove
        :type family_type_name: str
        """

        filtered_list = []
        removed_type= False
        for family_type_data_storage in self.family_type_data_storage:
            if family_type_data_storage.family_type_name !=  family_type_name:
                filtered_list.append(family_type_data_storage)
            else:
                # remove the family type data storage instance
                removed_type = True
                pass

        # replace the old list with the filtered list
        self._family_type_data_storage = filtered_list
        return removed_type


    def get_catalogue_file_data(self, parameter_names=None):
        """
        Get the catalogue file data for the family type data storage manager

        :param parameter_names: list of parameter names to include in the catalogue file data and the order in which they should be included.
        If None, all parameters will be included in the order they are stored in the family type data storage instances.
        :type parameter_names: [str]
        :return: the catalogue file data for the family type data storage manager
        :rtype: [[str]]
        """

        catalogue_file_data = []

        # loop over type data storage instances and get the catalogue file data
        for family_type_data_storage in self.family_type_data_storage:
            catalogue_file_data.append(
                family_type_data_storage.get_catalogue_file_data(parameter_names)
            )


        return catalogue_file_data


    def get_catalogue_file_header_row(self, parameter_names=None):
        """
        Get the header row for the catalogue file

        :param parameter_names: list of parameter names to include in the catalogue file data and the order in which they should be included.
        If None, all parameters will be included in the order they are stored in the family type data storage instances.
        :type parameter_names: [str]
        :return: the header row for the catalogue file
        :rtype: [str]
        """

        header_row = []

        # loop over type data storage instances and get the header row based on the first instance
        for family_type_data_storage in self.family_type_data_storage:
            header_row = family_type_data_storage.get_catalogue_file_header_row(parameter_names)
            break

        return header_row


    def get_report_data(self, project_name=None):
        """
        Get the report data for the family type data storage manager

        :return: the report data for the family type data storage manager
        :rtype: [[str]]
        """

        report_data = []

        for family_type_data_storage in self.family_type_data_storage:
            type_data = family_type_data_storage.get_report_data()
            for fam in type_data:
                report_data.append(fam)


        return report_data


    def get_difference(self, other):
        """
        Get the difference between two family type data storage managers

        Return format is a list of lists, where each list contains the following:
        - Family name
        - Family category
        - family exists in library
        - Family type name
        - Family type exists in Library
        - Parameter name
        - Parameter exists in library
        - Parameter difference

        An entry in the list is only returned if there is a difference between the two family type data storage managers

        :param other: the other family type data storage manager
        :type other: FamilyTypeDataStorageManager

        :return: the difference between the two family type data storage managers
        :rtype: [[str]]
        """

        if not isinstance(other, FamilyTypeDataStorageManager):
            raise ValueError(
                "other must be an instance of FamilyTypeDataStorageManager"
            )

        difference = []

        for family_type_data_storage in self.family_type_data_storage:

            # flag to check whether type (by name only exists in other)
            type_data_storage_has_match = False

            # loop over storage types from library to find a match
            for other_family_type_data_storage in other.family_type_data_storage:
                if family_type_data_storage.is_match_by_names_and_category(
                    other_family_type_data_storage
                ):
                    # set a flag for match
                    type_data_storage_has_match = True
                    # get the differences between the two family type data storage instances
                    diff_entries = (
                        family_type_data_storage.get_comparison_report_parameter_values(
                            other_family_type_data_storage
                        )
                    )
                    # if there are differences, add them to the difference list
                    for entry in diff_entries:
                        difference.append(entry)
                    break
                else:
                    continue

            if not type_data_storage_has_match:
                difference.append(
                    [
                        family_type_data_storage.family_name,
                        family_type_data_storage.root_category_path,
                        "Found match in library",
                        family_type_data_storage.family_type_name,
                        "No match in library",
                    ]
                )

        return difference


    def get_differences_report_header_row(self):
        """
        get the header row for the differences report

        :return: the header row for the differences report
        :rtype: [str]
        """

        return [
            "Family name",
            "Family category",
            "Family exists in library",
            "Family type name",
            "Family type exists in Library",
            "Parameter name",
            "Parameter exists in library",
            "Parameter difference (Project -> Library)",
        ]


    def remove_ghost_types(self, type_names):
        """
        Remove ghost types from the family type data storage manager
        Ghost types are types that are not present in the family file but listed in the part atom export from a Family class instance

        Any type that is not in the type_names list is considered a ghost type

        :param type_names: list of type names which are present in the document (project file) for this family
        :type type_names: [str]
        """

        filtered_list = []
        for family_type_data_storage in self.family_type_data_storage:
            if family_type_data_storage.family_type_name in type_names:
                filtered_list.append(family_type_data_storage)
            else:
                # remove the family type data storage instance
                pass

        # replace the old list with the filtered list
        self._family_type_data_storage = filtered_list



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family warnings data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData

# from duHast.Utilities import Utility as util
from duHast.Revit.Warnings import warnings as rWarn
from duHast.Revit.Warnings.Data.Objects.warnings_data_storage import (
    FamilyWarningsDataStorage,
)

# import Autodesk
# import Autodesk.Revit.DB as rdb

# WARNING_TEXT = "warningText"
# WARNING_GUID = "warningGUID"
# WARNING_RELATED_IDS = "warningRelatedIds"
# WARNING_OTHER_IDS = "warningOtherIds"


class WarningsData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None):
        """
        Constructor for warnings data class.

        :param root_path: root path for data
        :type root_path: str
        :param root_category_path: root category path for data
        :type root_category_path: str
        """

        super(WarningsData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
        )

    def process(self, doc):

        # make sure to get a value for the file path which is not empty if the document has not been saved
        saved_file_name = "-"
        if doc.PathName != "":
            saved_file_name = doc.PathName

        # get all warnings in document
        warnings = rWarn.get_warnings(doc)
        # loop over warnings and extract data
        for warning in warnings:
            # check for a guid
            war_guid = ""
            try:
                war_guid = warning.GetFailureDefinitionId().Guid
            except Exception as e:
                pass
            # warning text
            war_text = ""
            try:
                war_text = warning.GetDescriptionText()
            except Exception as e:
                pass
            # affected element ids
            war_element_ids_as_integer = []
            try:
                for el in warning.GetFailingElements():
                    war_element_ids_as_integer.append(el.IntegerValue)
            except Exception as e:
                pass
            # other element ids
            war_other_element_ids_as_integer = []
            try:
                for el in warning.GetAdditionalElements():
                    war_other_element_ids_as_integer.append(el.IntegerValue)
            except Exception as e:
                pass

            # build data
            storage = FamilyWarningsDataStorage(
                root_name_path=self.root_path,
                root_category_path=self.root_category_path,
                family_name=self._strip_file_extension(doc.Title),
                family_file_path=saved_file_name,
                warning_text=war_text,
                warning_guid=war_guid,
                warning_related_ids=war_element_ids_as_integer,
                warning_other_ids=war_other_element_ids_as_integer,
            )
            self.add_data(storage_instance=storage)

        # check if any shared parameter was found
        if len(self.data) == 0:
            # add message no warnings found
            # build data
            storage = FamilyWarningsDataStorage(
                root_name_path=self.root_path,
                root_category_path=self.root_category_path,
                family_name=self._strip_file_extension(doc.Title),
                family_file_path=saved_file_name,
                warning_text="No warnings present in family.",
                warning_guid="",
                warning_related_ids=[],
                warning_other_ids=[],
            )
            self.add_data(storage_instance=storage)

    def get_data(self):
        return self.data

    def add_data(self, storage_instance):
        if isinstance(storage_instance, FamilyWarningsDataStorage):
            self.data.append(storage_instance)
        else:
            raise ValueError(
                "storage instance must be an instance of FamilyWarningsDataStorage"
            )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit grids workset modifier functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import FilteredElementCollector, Grid

from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import worksets as rWork
from duHast.Revit.Grids import grids as rGrid


def modify_grid_worksets_default(doc, worksetRules):
    """
    Workset modifier method. Moves all grids to one workset.

    Rules format:

    ['Model name', [[workset modifier method,[worksetName]]]]

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param worksetRules: The workset rules in the specified format.
    :type worksetRules: List

    :return: A result object.
    :rtype: Type
    """

    gridsResults = res.Result()
    collectorGrids = FilteredElementCollector(doc).OfClass(Grid)
    for rule in worksetRules:
        for defaultWorksetName in rule:
            grids = rWork.modify_element_workset(
                doc, defaultWorksetName, collectorGrids, "grids"
            )
            gridsResults.update(grids)
    return gridsResults


def modify_grid_worksets_by_type_name(doc, worksetRules):
    """
    Workset modifier method. Moves grids matching a type condition to a particular workset.

    Default workset type rules example:

    [['model name', [[ModifyGridWorkSetsByTypeName, [['workset name', util.ConDoesNotEqual, 'grid type name'],['workset name', util.ConDoesEqual, 'grid type name']]]]]]

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param worksetRules: The workset rules in the specified format.
    :type worksetRules: List

    :return: A result object.
    :rtype: Type
    """

    gridsResults = res.Result()
    # loop over grid type filter and address one at the time
    # get all grids matching type name filter
    for (
        defaultWorksetName,
        typeNameCondition,
        typeName,
    ) in worksetRules:
        # get the grid type id from the type name
        typeId = rGrid.get_grid_type_id_by_name(doc, typeName)
        collectorGrids = (
            FilteredElementCollector(doc)
            .OfClass(Grid)
            .Where(lambda e: typeNameCondition(e.GetTypeId(), typeId))
        )
        grids = rWork.modify_element_workset(
            doc, defaultWorksetName, collectorGrids, "grids"
        )
        gridsResults.update(grids)
    return gridsResults


def modify_grid_worksets_by_parameter_value(doc, worksetRules):
    """
    Workset modifier method. Moves grids matching parameter condition to a particular workset

    Default workset type rules example:

    [['model name', [[ModifyGridWorkSetsByParameterValue, [['workset name', util.ConTwoStartWithOne, 'Name', 'name starts with value']]]]]]

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param worksetRules: _description_
    :type worksetRules: _type_

    :return: returns a result object
    :rtype: _type_
    """

    gridsResults = res.Result()
    # loop over grid parameter filter and address one at the time
    # get all grids matching filter
    for defaultWorksetName, paraCondition, paraName, conditionValue in worksetRules:
        collectorGrids = (
            FilteredElementCollector(doc)
            .OfClass(Grid)
            .Where(
                lambda e: rGrid.grid_check_parameter_value(
                    e, paraName, paraCondition, conditionValue
                )
            )
        )
        grids = rWork.modify_element_workset(
            doc, defaultWorksetName, collectorGrids, "grids"
        )
        gridsResults.update(grids)
    return gridsResults


def modify_grids_worksets(doc, revitFileName, worksetRules):
    """
    Modifies worksets of grids as per workset rules format.

    Default workset type rules example:

    [['model name', [[ModifyGridWorkSetsDefault, [['default workset name']  # There should only be one per model]]]]]

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revitFileName: _description_
    :type revitFileName: _type_
    :param worksetRules: _description_
    :type worksetRules: _type_

    :return: returns a result object
    :return: Result class instance.

        - result.status: View deletion status returned in result.status. False if an exception occurred, otherwise True.
        - result.message: will contain the fully qualified file path of the exported file
        - result.result: will be an empty list

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message

    :rtype: :class:`.Result`
    """

    gridsResults = res.Result()
    foundMatch = False
    for fileName, worksetModifierList in worksetRules:
        if revitFileName.startswith(fileName):
            foundMatch = True
            for worksetModifier, rules in worksetModifierList:
                grids = worksetModifier(doc, rules)
                gridsResults.update(grids)
            break
    if foundMatch == False:
        gridsResults.update_sep(False, "No grid rules found for file: " + revitFileName)
    return gridsResults



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit purge utility functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#


from Autodesk.Revit.DB import BuiltInParameter, ElementId

from duHast.Revit.Common import parameter_get_utils as rParaGet


def build_category_dictionary(doc, element_ids):
    """
    Builds a dictionary from elementId s past in.
    Dictionary key is the element category and values are all the elements of that category.
    If no category can be found the key 'invalid category' will be used.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_ids: List of element id of which to build the dictionary from.
    :type element_ids: list of AutoDesk.Revit.DB.ElementId
    :return: Dictionary key is the element category and values are all the elements of that category.
    :rtype: dictionary, key is string, value is list of AutoDesk.Revit.DB.Element
    """

    dic = {}
    for element_id in element_ids:
        try:
            el = doc.GetElement(element_id)
            try:
                if el.Category.Name in dic:
                    dic[el.Category.Name].append(el)
                else:
                    dic[el.Category.Name] = [el]
            except:
                if "invalid category" in dic:
                    dic["invalid category"].append(el)
                else:
                    dic["invalid category"] = [el]
        except:
            if "invalid element" in dic:
                dic["invalid element"].append(el)
            else:
                dic["invalid element"] = [el]
    return dic


def check_whether_dependent_elements_are_multiple_orphaned_legend_components(
    doc, element_ids
):
    """
    Check if element are orphaned legend components

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_ids: List of elements to check
    :type element_ids: list of AutoDesk.Revit.DB.ElementId
    :return: True if all but one element are orphaned legend components.
    :rtype: bool
    """

    flag = True
    category_name = "Legend Components"
    # build dependent type dictionary
    # check whether dictionary is made of
    #   1 entry for type
    #   multiple entries for legend components
    #   no other entry
    # if so: check whether any of the legend component entry has a valid view id
    #   if none has return true, otherwise return false
    dic = build_category_dictionary(doc, element_ids)
    # check if dictionary has legend component key first up
    if category_name in dic:
        # if so check number of keys and length of elements per key
        if len(dic.keys()) == 2 and len(dic[category_name]) == len(element_ids) - 1:
            # this should be the only code path returning true...
            for value in dic[category_name]:
                if value.OwnerViewId != ElementId.InvalidElementId:
                    flag = False
                    break
        else:
            flag = False
    else:
        flag = False
    return flag


def filter_out_warnings(doc, dependent_elements):
    """
    Attempts to filter out any warnings from ids supplied by checking the workset name
    of each element for 'Reviewable Warnings'

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param dependent_elements: List of elements to check.
    :type dependent_elements: list of AutoDesk.Revit.DB.Element
    :return: A list of elements id where the workset name of the element is not 'Reviewable Warnings'
    :rtype: list of AutoDesk.Revit.DB.Element
    """

    ids = []
    for id in dependent_elements:
        el = doc.GetElement(id)
        p_value = rParaGet.get_built_in_parameter_value(
            el, BuiltInParameter.ELEM_PARTITION_PARAM, rParaGet.get_parameter_value
        )
        if p_value != "Reviewable Warnings":
            ids.append(id)
    return ids


def has_dependent_elements(
    doc, el, filter=None, threshold=2  # type: int
):
    """
    Checks whether an element has dependent elements.
    The dependent elements are collected via Element.GetDependentElements(filter).
    This also includes a check as to whether elements returned as dependent are orphaned. (for lack of better words)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param el: The element to be checked for dependent elements.
    :type el: AutoDesk.Revit.DB.Element
    :param filter: What type of dependent elements to filter, defaults to None which will return all dependent elements
    :type filter: Autodesk.Revit.DB.ElementFilter , optional
    :param threshold: The number of how many dependant elements an element can have but still be considered not used, defaults to 2
    :type threshold: int, optional
    :return: returns 0 for no dependent elements, 1, for other elements depend on it, -1 if an exception occurred
    :rtype: int
    """

    value = 0  # 0: no dependent Elements, 1: has dependent elements, -1 an exception occurred
    try:
        dependentElements = el.GetDependentElements(filter)
        # remove any warnings from dependent elements
        dependentElements = filter_out_warnings(doc, dependentElements)
        # check if dependent elements pass threshold value
        if (len(dependentElements)) > threshold:
            # there appear to be situations where dependent elements are multiple (orphaned?) legend components only
            # or warnings belonging to a type (same type mark ...)
            # these are legend components with an invalid OwnerViewId, check whether this is the case...
            if (
                check_whether_dependent_elements_are_multiple_orphaned_legend_components(
                    doc, dependentElements
                )
                == False
            ):
                value = 1
    except Exception as e:
        value = -1
    return value


def get_used_unused_type_ids(
    doc,
    type_id_getter,
    use_type=0,  # type: int
    threshold=2,  # type: int
):
    """
    Gets either the used or not used type Ids provided by typeIdGetter.
    Whether the used or unused type ids depends on the useType value.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_id_getter: Function returning type ids
    :type type_id_getter: list of Autodesk.Revit.DB.ElementId
    :param use_type: 0, no dependent elements; 1: has dependent elements, defaults to 0
    :type use_type: int, optional
    :param threshold: The number of how many dependant elements an element can have but still be considered not used, defaults to 2
    :type threshold: int, optional
    :return: A list of either all used or unused element ids. Depends on useType.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get all types elements available
    all_type_ids = type_id_getter(doc)
    ids = []
    for type_id in all_type_ids:
        type = doc.GetElement(type_id)
        has_dependents = has_dependent_elements(doc, type, None, threshold)
        if has_dependents == use_type:
            ids.append(type_id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions for benchmarking:. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- two values
- whether a text value starts or does not start with a given text value

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import time
import functools


def measure_time_wrapper(func, logger=None, measure=True):
    # Get the function name for the print statement
    func_name = func.__name__

    @functools.wraps(func)
    def wrapper_func(*args, **kwargs):
        # Get the start time
        start = time.time()
        # Execute the function
        result = func(*args, **kwargs)
        # Get the end time
        end = time.time()
        # Calculate the elapsed time
        elapsed = round(float(end - start), 5)
        message = "Elapsed time for {} is : {}".format(func_name, elapsed)
        if logger:
            # Log the elapsed time
            logger.info(message)
        else:
            # Print the elapsed time
            print(message)
        # Return the result of the function
        return result

    if measure:
        return wrapper_func
    else:
        return func


def add_measure_time(func, should_output=True):
    """
    Decorator function to measure the time a function takes to execute and then print it to the console.

    Use like
    @measure_time(should_output = show_benchmarking_output_bool)
    def function_to_measure():


    :param should_output: Whether the elapsed time should be printed to the console
    :type should_output: bool
    """

    if should_output:
        return measure_time_wrapper(func, measure=should_output)
    else:
        return func



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around exporting from Revit to data objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import Element, StorageType

from duHast.Revit.Common.parameter_get_utils import (
    getter_double_as_double_converted_to_metric,
    getter_int_as_int,
    getter_string_as_UTF8_string,
    getter_element_id_as_element_int,
    getter_none,
    get_all_parameters_and_values_wit_custom_getters,
    get_built_in_parameter_value,
)
from duHast.Revit.Common.design_set_options import get_design_set_option_info
from duHast.Utilities.utility import encode_utf8
from duHast.Data.Objects.Collectors.Properties.data_property import DataProperty
from duHast.Revit.Common.phases import (
    get_phase_id_created,
    get_phase_id_demolished,
    get_phase_name_by_id,
)
from duHast.Revit.Common.Objects.design_set_property_names import DesignSetPropertyNames
from duHast.Data.Objects.Collectors.Properties.data_phasing import DataPhasing
from duHast.Data.Objects.Collectors.Properties.data_revit_model import DataRevitModel
from duHast.Data.Objects.Collectors.Properties.data_level import DataLevel
from duHast.Data.Objects.Collectors.Properties.data_type_properties import DataTypeProperties
from duHast.Data.Objects.Collectors.Properties.data_instance_properties import (
    DataInstanceProperties,
)
from duHast.Data.Objects.Collectors.Properties.data_design_set_option import DataDesignSetOption


def get_element_properties(element):
    """
    Converts properties on a Revit element into data_property instances

    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: A list of data_property instances
    :rtype: [:class:`.DataProperty`]
    """
    # custom parameter value getters
    value_getter = {
        StorageType.Double: getter_double_as_double_converted_to_metric,
        StorageType.Integer: getter_int_as_int,
        StorageType.String: getter_string_as_UTF8_string,  # encode ass utf 8 just in case
        StorageType.ElementId: getter_element_id_as_element_int,  # needs to be an integer for JSON encoding
        str(None): getter_none,
    }

    # convert properties from Revit into data_property instances
    properties = []
    for revit_prop in get_all_parameters_and_values_wit_custom_getters(
        element, value_getter
    ).items():
        data_p = DataProperty()
        data_p.name = revit_prop[0]
        data_p.value = revit_prop[1]
        properties.append(data_p)

    return properties


def get_type_properties(doc, element):
    """
    Returns a populated data type property instance.

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: A data type properties instance
    :rtype: [:class:`.DataTypeProperties`]
    """
    type_p = DataTypeProperties()

    # get element properties
    type_p.id = element.GetTypeId().IntegerValue
    type_p.name = encode_utf8(Element.Name.GetValue(element))
    element_type = doc.GetElement(element.GetTypeId())

    # get the elements type properties
    type_properties = get_element_properties(element=element_type)
    type_p.properties = type_properties

    return type_p


def get_instance_properties(element):
    """
    Returns a populated data instance property instance.

    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: A data instance properties instance
    :rtype: [:class:`.DataInstanceProperties`]
    """
    instance_p = DataInstanceProperties()

    # get instance properties
    instance_p.id = element.Id.IntegerValue
    instance_properties = get_element_properties(element=element)
    instance_p.properties = instance_properties

    return instance_p


def get_level_data(doc, element, built_in_parameter_def):
    """
    Returns a populated data level instance.

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element
    :param built_in_parameter_def: The build in parameter containing the offset from level
    :type built_in_parameter_def: Autodesk.Revit.DB.BuiltInParameter

    :return: A data level instance
    :rtype: [:class:`.DataLevel`]
    """

    level_d = DataLevel()

    # get level properties
    level_d.name = encode_utf8(Element.Name.GetValue(doc.GetElement(element.LevelId)))
    level_d.id = element.LevelId.IntegerValue
    level_d.offset_from_level = get_built_in_parameter_value(
        element=element, built_in_parameter_def=built_in_parameter_def
    )  # offset from level

    return level_d


def get_design_set_data(doc, element):
    """
    Returns a populated data design set option instance.

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: A data design set instance
    :rtype: [:class:`.DataDesignSetOption`]
    """

    design_set_d = DataDesignSetOption()

    # get design set data
    design_set_data = get_design_set_option_info(doc, element)
    design_set_d.option_name = design_set_data[
        DesignSetPropertyNames.DESIGN_OPTION_NAME
    ]
    design_set_d.set_name = design_set_data[DesignSetPropertyNames.DESIGN_SET_NAME]
    design_set_d.is_primary = design_set_data[
        DesignSetPropertyNames.DESIGN_OPTION_IS_PRIMARY
    ]

    return design_set_d


def get_phasing_data(doc, element):
    """
    Returns a data phase instance with names for created and demolished populated.

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: A data phase instance
    :rtype:  [:class:`.DataPhasing`]
    """

    phase_d = DataPhasing()
    phase_d.created = get_phase_name_by_id(
        doc=doc, phase_id=get_phase_id_created(element=element)
    )
    phase_d.demolished = get_phase_name_by_id(
        doc=doc, phase_id=get_phase_id_demolished(element=element)
    )

    return phase_d


def get_model_data(doc):
    """
    Returns a revit_model instance with the name populated. If detached and not saved yet it will be named "Detached Model"

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A data revit model instance
    :rtype:  [:class:`.DataRevitModel`]
    """

    model_d = DataRevitModel()
    # set a default value in case the model is detached
    model_name = "Detached Model"
    # get the model name
    if doc.IsDetached == False:
        model_name = doc.Title
    model_d.name = model_name
    return model_d



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit line line styles helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from Autodesk.Revit.DB import BuiltInCategory, ElementId, GraphicsStyleType
from duHast.Revit.Common import delete as rDel


def delete_line_styles_starts_with(doc, starts_with):
    """
    Deletes all line styles where the name starts with provided string

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param starts_with: Filter: style name needs to start with this string to be deleted.
    :type starts_with: str
    :return: Result class instance.

        - .status True if all views where deleted. Otherwise False.
        - .message will contain deletion status.

    :rtype: :class:`.Result`
    """

    lc = doc.Settings.Categories[BuiltInCategory.OST_Lines]
    ids = list(c.Id for c in lc.SubCategories if c.Name.StartsWith(starts_with)).ToList[
        ElementId
    ]()
    result = rDel.delete_by_element_ids(
        doc,
        ids,
        "Delete line styles where name starts with: " + str(starts_with),
        "line styles starting with: " + str(starts_with),
    )
    return result


def get_all_line_style_ids(doc):
    """
    Gets all line styles ids in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of all line style ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    lc = doc.Settings.Categories[BuiltInCategory.OST_Lines]
    ids = list(c.Id for c in lc.SubCategories).ToList[ElementId]()
    return ids


def get_all_graphics_style_ids_by_line_style_id(doc):
    """
    Gets all graphics style ids by line style id in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A dictionary of all line style ids as integer values with their corresponding graphics style ids as integer values.
    :rtype: dict
    """

    return_value = {}
    cat = doc.Settings.Categories.get_Item(BuiltInCategory.OST_Lines)
    sub_cats = cat.SubCategories
    for line_style in sub_cats:
        graphic_style = line_style.GetGraphicsStyle(GraphicsStyleType.Projection)
        return_value[line_style.Id.IntegerValue] = graphic_style.Id.IntegerValue
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit door export to DATA class functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    BoundingBoxXYZ,
    BuiltInParameter,
    Options,
)


# from duHast.APISamples.Ceilings.Geometry import Geometry
from duHast.Revit.Doors.doors import get_door_instances
from duHast.Data.Objects.Collectors import data_door as dDoor
from duHast.Revit.Common.Geometry.to_data_conversion import (
    convert_bounding_box_to_flattened_2d_points,
    convert_xyz_in_data_geometry_polygons,
)

from duHast.Revit.Common.Geometry.solids import get_bounding_box_from_family_geometry
from duHast.Revit.Exports.export_data import (
    get_level_data,
    get_phasing_data,
    get_model_data,
    get_instance_properties,
    get_type_properties,
    get_design_set_data,
)


def populate_data_door_object(doc, revit_door):
    """
    Returns a custom ceiling data objects populated with some data from the revit model ceiling past in.

    - door id
    - door type name
    - door mark
    - door type mark
    - door level name
    - door level id
    - door offset from level

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_door: A revit door instance.
    :type revit_door: Autodesk.Revit.DB.FamilyInstance

    :return: A data door object instance.
    :rtype: :class:`.DataDoor`
    """

    # set up data class object
    data_door = dDoor.DataDoor()
    # get door bounding box
    # set a default option
    opts = Options()

    door_bounding_box = get_bounding_box_from_family_geometry(
        revit_door.get_Geometry(opts)
    )
    # only export door data if 3D geometry is available
    # if no geometry is available, return None
    # some families of type door do not have any 3D geometry... ignore those for now
    if isinstance(door_bounding_box, BoundingBoxXYZ) == False:
        return None
    revit_door_data_geometry = convert_bounding_box_to_flattened_2d_points(
        door_bounding_box
    )

    if len(revit_door_data_geometry.outer_loop) > 0:
        door_point_groups_as_doubles = []
        data_geo_converted = convert_xyz_in_data_geometry_polygons(
            doc, revit_door_data_geometry
        )
        door_point_groups_as_doubles.append(data_geo_converted)
        data_door.polygon = door_point_groups_as_doubles

        # get design set data
        design_set = get_design_set_data(doc=doc, element=revit_door)
        data_door.design_set_and_option = design_set

        # get type properties
        type_props = get_type_properties(doc=doc, element=revit_door)
        data_door.type_properties = type_props

        # get instance properties
        instance_props = get_instance_properties(revit_door)
        data_door.instance_properties = instance_props

        # get level properties
        level = get_level_data(
            doc=doc,
            element=revit_door,
            built_in_parameter_def=BuiltInParameter.ASSOCIATED_LEVEL_OFFSET,  # TODO: check this parameter
        )
        data_door.level = level

        # get the model name
        model = get_model_data(doc=doc)
        data_door.revit_model = model

        # get phasing information
        phase = get_phasing_data(doc=doc, element=revit_door)
        data_door.phasing = phase

        return data_door
    else:
        return None


def get_all_door_data(doc, filter_family_names=[]):
    """
    Gets a list of door data objects for each door element in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of data door instances.
    :rtype: list of :class:`.DoorCeiling`
    """

    all_door_data = []
    doors = get_door_instances(doc)
    for door in doors:
        # check if door is in filter list
        if len(filter_family_names) > 0:
            if door.Symbol.FamilyName not in filter_family_names:
                continue
        door_data = populate_data_door_object(doc, door)
        if door_data is not None:
            all_door_data.append(door_data)
    return all_door_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing comparison reporting of family reports.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This mimicks an outer join between any number of reports. 

The reports are compared based on the family name, type name, parameter name:, 


family Name,    Type name,  parmater name,  report 1,   report 2,   report 3,   ...
family 1,       type 1,     parameter 1,    value 1,   N/A,        value 1,     ...
family 2,       type 2,     parameter 2,    N/A,       value 2,    value 3,     ...
family 3,       type 3,     parameter 3,    value 3,       N?A      N/A         ...


N/A = not available in that report
"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_csv import read_csv_file
from duHast.Utilities.files_io import get_file_name_without_ext


def _build_report_dict(reports):
    # build a dictionary of the report
    report_dic = {}
    for report in reports:
        for report_name, report_content in report.items():
            for row in report_content [1:]:
                family_name = row[1]
                family_category = row[2]
                type_name = row[3]
                parameter_name = row[6]
                value = row[10]
                report_dic[(family_name, family_category, type_name, parameter_name)] = value
    
    return report_dic

        
def outer_join_report_a_vs_b(report_a, report_b):

    # this will compare initial a vs b and b vs a as well
    # the return will be a list of lists:
    # [ family Name,    Type name,  parmater name,  value report 1,   value report 2]
    
    # build a dictionary for each report:
    report_a_dict = _build_report_dict(report_a)
    report_b_dict = _build_report_dict(report_b)

    # Get all unique keys from both reports
    all_keys = set(report_a_dict.keys()).union(set(report_b_dict.keys()))
    
    # Perform the outer join
    outer_join = {}
    for key in all_keys:
        outer_join[key] = [report_a_dict.get(key), report_a_dict.get(key)]
    
    return outer_join


def compare_outer_join_result_with_report(report, outer_join_result):
    
    # build a dictionary from the report:
    report_dict = _build_report_dict(report)

    # Get all unique keys from both reports
    all_keys = set(report_dict.keys()).union(set(outer_join_result.keys()))

    # get the number of paddings required for the outer join result
    n_a_value = "N/A"
    padding_length = 2
    for key in outer_join_result:
        padding_length = len(outer_join_result[key])
        break

    # Perform the outer join:
    # if there is a match for a key in both dictionaries, then the value of the report need to be added to the outer join result for that key, which is a tuple of values
    # if there is no match for a key from the outer join result in the report, then N?A needs to be added to the outer join result for that key
    # if there is no match for a key from the report in the outer join result, then the key needs to be added to the outer join result and the value of the report need to be added
    # but the value needs to be padded with N/A for the other reports

    for key in all_keys:
        if key in report_dict:
            if key in outer_join_result:
                outer_join_result[key] = outer_join_result[key] + [report_dict[key],]
            else:
                outer_join_result[key] = [n_a_value for _ in range(padding_length)] + [report_dict[key],]
        else:
            outer_join_result[key] = outer_join_result[key] + [n_a_value,]
    
    return outer_join_result
   

def read_report(file_path):

    return_value = Result()
    data = {}

    try:
        # get the report name
        report_name = get_file_name_without_ext(file_path)
        # read the report
        read_result = read_csv_file(file_path)
        # check if the read was successful
        if read_result.status:
            data[report_name] = read_result.result
            return_value.append_message("Report read successfully: {}".format(report_name))
            return_value.result.append(data)
        else:
            # something went wrong...
            return_value.update_sep(False, read_result.message)
            return return_value
    except Exception as e:
        return_value.update_sep(False, str(e))
    return return_value


def compare_family_reports_outer_join(file_paths):

    return_value = Result()
    # make sure there at least two reports
    if len(file_paths) < 2:
        raise ValueError("At least two reports are required")
    
    # read reports into memory
    reports = []
    for file_path in file_paths:
        # get the report
        read_report_result = read_report(file_path)
        # check if reead ok, otherwise get out
        if not read_report_result.status:
            return read_report_result
        reports.append(read_report_result.result)
    
    #l og updates
    return_value.append_message("read {} reports.".format(len(reports)))
    
    # compare reports
    # do the initial report comparison
    outer_join_results = outer_join_report_a_vs_b(report_a= reports[0], report_b= reports[1])

    # get out if there are only 2 reports to compare
    if len(reports) == 2:
        return_value.result = outer_join_results
        return return_value
    
    # compare the outer join results with the remaining reports
    for report in reports[2:]:
        outer_join_results = compare_outer_join_result_with_report(report, outer_join_results)
    
    return_value.result = outer_join_results
    return  return_value

    



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains utility function(s) for Revit link reports. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import BuiltInParameter, Element, ModelPathUtils

from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Links.links import (
    get_all_revit_link_instances,
    get_revit_link_type_from_instance,
)


def get_revit_link_type_data(doc, revit_link_type):
    """
    Gets Revit Link Type data for reporting.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_link_type: The link type of which to get the data from.
    :type revit_link_type: Autodesk.Revit.DB.RevitLinkType
    :return: A list of string
    :rtype: list str
    """

    # default values
    model_path = "unknown"
    is_loaded = False
    is_from_local_path = False
    path_type = "unknown"
    is_loaded = revit_link_type.IsLoaded(doc, revit_link_type.Id)
    is_from_local_path = revit_link_type.IsFromLocalPath()

    # check external file reference
    # this will fail if link is to a cloud file
    ex_file_ref = None
    try:
        ex_file_ref = revit_link_type.GetExternalFileReference()
    except:
        # set path type to cloud file
        path_type = "cloud file"

    # check if we got a valid external file reference
    if ex_file_ref == None:
        # try to get the external file reference from the first external resource reference instead
        # since this is not a local file server link
        ex_resource_references = revit_link_type.GetExternalResourceReferences()
        # the path value will be empty if the link is not loaded
        for ref in ex_resource_references:
            # these are c# key value pairs
            if ref.Value.InSessionPath == "":
                model_path = "unknown since link is unloaded"
            else:
                model_path = ref.Value.InSessionPath
            break
    else:
        # local file server link
        if ex_file_ref.IsValidExternalFileReference(ex_file_ref):
            model_path = ModelPathUtils.ConvertModelPathToUserVisiblePath(
                ex_file_ref.GetPath()
            )
            path_type = ex_file_ref.PathType.ToString()

    # get the workset of the link type (this can bew different to the workset of the link instance)
    ws_parameter = revit_link_type.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)

    data = [
        Element.Name.GetValue(revit_link_type),
        str(is_loaded),
        str(ws_parameter.AsValueString()),
        str(is_from_local_path),
        path_type,
        model_path,
    ]

    return data


def get_revit_link_report_data(doc, revit_file_path):
    """
    Gets link data ready for being printed to file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The file hostname, which is added to data returned.
    :type revit_file_path: str
    :return: list of list of revit link properties.
    :rtype: list of list of str
    """

    data = []
    collector = get_all_revit_link_instances(doc)
    for c in collector:
        # get the workset
        ws_parameter = c.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)
        # get the design option
        do_parameter = c.get_Parameter(BuiltInParameter.DESIGN_OPTION_ID)
        # get whether link is shared or not (only works when link is loaded)
        if "<Not Shared>" in c.Name:
            l_s = False
        else:
            l_s = True
        # get shared location name ( needs to be in try catch in case file is unloaded)
        link_location_name = "unknown"
        try:
            link_location_name = c.GetLinkDocument().ActiveProjectLocation.Name
        except Exception:
            pass
        link_type = get_revit_link_type_from_instance(doc, c)
        # get link type data ( this is another list)
        link_type_data = get_revit_link_type_data(doc, link_type)
        # add other data
        link_type_data_combined = [
            revit_file_path,
            str(c.Id),
            str(l_s),
            link_location_name,
            rParaGet.get_parameter_value(ws_parameter),
            rParaGet.get_parameter_value(do_parameter),
        ]
        # splice the two lists together
        link_type_data_combined[2:] = link_type_data
        # add link to value to be returned
        data.append(link_type_data_combined)
    return data
